// Copyright (C) 2000 Dominic Letourneau (doumdi@yahoo.com)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "Smooth.h"
#include "Object.h"
#include "ObjectRef.h"
#include "Exception.h"
#include "Kernel.h"
#include <math.h>

using namespace std;

NODE_INFO(Smooth,"Image", "KERNEL_IN:IMAGE_IN", "IMAGE_OUT", "")

Smooth::Smooth(string nodeName, ParameterSet params)
: Node(nodeName, params)
{
  // Add the node's input
  m_KERNEL_ID = addInput("KERNEL_IN");
  m_IMAGE_ID = addInput("IMAGE_IN");
  
  // Add the node's output
  addOutput("IMAGE_OUT");
  
}

Smooth::~Smooth()
{
  return;
}

GreyScale8Image* Smooth::convolve(GreyScale8Image &im, Kernel &kernel)
{

  int kernel_size = kernel.get_size();
  int image_width = im.get_width();
  int image_height = im.get_height();

  if (kernel_size % 2 != 1) {
    throw new NodeException (this, "The convolution kernel must be a square and have odd dimensions.",__FILE__,__LINE__);
  }
  
  //creating output image.
  GreyScale8Image *out = new GreyScale8Image(image_width,image_height);
  
  //data pointers
  int *k_ptr_base = kernel.get_data();
  int *k_ptr = NULL;

  unsigned char *im_ptr_base = im.get_data();
  unsigned char *im_ptr = NULL;

  unsigned char *out_ptr = out->get_data();
 
  
  int half = kernel_size / 2;
  int min_i;
  int max_i;
  int min_j;
  int max_j;
  int i,j,k,l;
  int sum;
  int k_sum;

  for (i = 0; i < image_height; i++) {

    min_i = max(-half,-i);
    max_i = min(half, image_height - 1 - i);

    for (j = 0; j < image_width; j++) {         
      min_j = max(-half, -j);
      max_j = min(half,  image_width - 1 - j);
      
      sum = 0;
      k_sum = 0;
      
      for(k = min_i; k <= max_i; k++) {
	for(l = min_j; l <= max_j; l++) {
	  int coef =  k_ptr_base[(k + half)*kernel_size + l + half];
	  sum += im_ptr_base[(i+k)*image_width + (j+l)] * coef;
	  k_sum += coef;
	   
	}
      }
      if (k_sum != 0) {
	sum /= k_sum;
      }
    
      if (sum > 0)
	*out_ptr = min(sum,255);
      else
	*out_ptr = 0;

      out_ptr++;
    }
    
    
  }
  return out;
}
	
ObjectRef Smooth::getOutput (int output_id, int count) {

  if (!hasOutput(output_id)) throw new NodeException (this, "Cannot getOutput id",__FILE__,__LINE__);
  
   
  if (count != processCount) {
    //We are updating our output only if needed
    
    try {
      
      //getting all data from our inputs.
      int OutputID = inputs[m_IMAGE_ID].outputID;
      ObjectRef im = inputs[m_IMAGE_ID].node->getOutput(OutputID, count);
      
      OutputID = inputs[m_KERNEL_ID].outputID;
      ObjectRef kernel = inputs[m_KERNEL_ID].node->getOutput(OutputID,count);
      
      GreyScale8Image *im_ptr = convolve(object_cast<GreyScale8Image>(im),object_cast<Kernel>(kernel));
      
      m_output = ObjectRef(im_ptr);

    } //end of try block
    catch (BaseException *e) {
      //Something weird happened
      //e->print();
      throw e->add(new NodeException (this,
				      string("error caught in Smooth::getOutput(int,int)"), 
				      __FILE__,__LINE__));
     }      
     
    //updating processCount
    processCount = count;                 
  }
  
  return (m_output);
  
} 
 
 
